class AugerConfig(object):
    """Modify configuration options in auger.yaml."""

    def __init__(self, ctx):
        super(AugerConfig, self).__init__()
        self.ctx = ctx

    def _with_auger_config(decorated):
        def wrapper(self, *args, **kwargs) :
            decorated(self, self.ctx.config, *args, **kwargs)
            self.ctx.config.write('auger')
        return wrapper

    @_with_auger_config
    def config(self, config, *args, **kwargs):
        config.set('auger', 'experiment/name',
            kwargs.get('experiment_name', ''))
        config.set('auger', 'dataset',
            kwargs.get('data_set_name', ''))
        config.set('auger', 'experiment/experiment_session_id',
            kwargs.get('experiment_session_id', ''))
        # model_type = kwargs.get('model_type', None)
        # if model_type:
        #     yaml['experiment']['metric'] = \
        #         'f1_macro' if model_type == 'classification' else 'r2'

    @_with_auger_config
    def set_data_set(self, config, data_set_name):
        config.set('auger', 'dataset', data_set_name)

    @_with_auger_config
    def set_experiment(self, config, experiment_name, experiment_session_id):
        config.set('auger', 'experiment/name', experiment_name)
        config.set('auger', 'experiment/experiment_session_id',
            experiment_session_id)



import os
import pytest
from a2ml.api.azure.a2ml import AzureA2ML
from a2ml.api.utils.context import Context


class TestAzureA2ML():

    @pytest.mark.skip(reason='make it work first')
    def test_init(self):
        print("Current directory: {}".format(os.getcwd()))
        fulldir=os.getcwd()+"/tests/test_app"
        os.chdir(fulldir)

    @pytest.mark.skip(reason='make it work first')
    def test_import_data(self):
        print("Current directory: {}".format(os.getcwd()))
        self.ctx = Context()
        print("Project name: {}".format(self.ctx.config.get('name')))
        a2ml=AzureA2ML(self.ctx.copy('azure'))
        a2ml.import_data()



from a2ml.api.utils.crud_runner import CRUDRunner
from a2ml.api.utils.error_handler import ErrorHandler

class A2MLModel(metaclass=ErrorHandler):

    def __init__(self, ctx, provider):
        super(A2MLModel, self).__init__()
        self.ctx = ctx
        self.runner = CRUDRunner(ctx, provider, 'model')

    def deploy(self, model_id, locally):
        return self.runner.execute('deploy', model_id, locally)

    def predict(self, filename, model_id, threshold, locally):
        return self.runner.execute('predict', filename, model_id, threshold, locally)

    def actual(self, filename, model_id):
        return self.runner.execute('actual', filename, model_id)



import click
from a2ml.api.a2ml_model import A2MLModel
from a2ml.api.utils.context import pass_context


@click.group('model', short_help='Model management')
@pass_context
def cmdl(ctx):
    """Model management"""
    ctx.setup_logger(format='')

@click.command('deploy', short_help='Deploy trained model.')
@click.argument('model-id', required=True, type=click.STRING)
@click.option('--locally', is_flag=True, default=False,
    help='Download and deploy trained model locally.')
@click.option('--provider', '-p', type=click.STRING, required=False,
    help='Cloud AutoML Provider.')
@pass_context
def deploy(ctx, provider, model_id, locally):
    """Deploy trained model."""
    A2MLModel(ctx, provider).deploy(model_id, locally)

@click.command('predict', short_help='Predict with deployed model.')
@click.argument('filename', required=True, type=click.STRING)
@click.option('--threshold', '-t', default=None, type=float,
    help='Threshold.')
@click.option('--model-id', '-m', type=click.STRING, required=True,
    help='Deployed model id.')
@click.option('--locally', is_flag=True, default=False,
    help='Predict locally using Docker image to run model.')
@click.option('--provider', '-p', type=click.STRING, required=False,
    help='Cloud AutoML Provider.')
@pass_context
def predict(ctx, provider, filename, model_id, threshold, locally):
    """Predict with deployed model."""
    A2MLModel(ctx, provider).predict(filename, model_id, threshold, locally)

@click.command('actual', short_help='Send actual values for deployed model. Needed for review and monitoring.')
@click.argument('filename', required=True, type=click.STRING)
@click.option('--model-id', '-m', type=click.STRING, required=True,
    help='Deployed model id.')
@click.option('--provider', '-p', type=click.STRING, required=False,
help='Cloud AutoML Provider.')
@pass_context
def actual(ctx, provider, filename, model_id):
    """Predict with deployed model."""
    A2MLModel(ctx, provider).actual(filename, model_id)


@pass_context
def add_commands(ctx):
    cmdl.add_command(deploy)
    cmdl.add_command(predict)
    cmdl.add_command(actual)


add_commands()



import os
import time
import json
import logging
import numpy as np
import pandas as pd
from azureml.core import Workspace
from azureml.core import Experiment
from azureml.core import diagnostic_log
from azureml.core.dataset import Dataset
from azureml.core.compute import AmlCompute
from azureml.core.compute import ComputeTarget
from azureml.train.automl import AutoMLConfig
from azureml.core.experiment import Experiment
from azureml.train.automl.run import AutoMLRun
# deployment
from azureml.core.model import InferenceConfig
from azureml.core.webservice import AciWebservice
from azureml.core.webservice import Webservice
from azureml.exceptions import WebserviceException
from azureml.core.model import Model

from a2ml.api import a2ml
from a2ml.api.utils.formatter import print_table
from a2ml.api.azure.decorators import error_handler
from a2ml.api.azure.exceptions import AzureException

logging.basicConfig(level=logging.INFO)


class AzureA2ML(object):
    def __init__(self, ctx):
        diagnostic_log().start_capture()
        self.ctx = ctx
        project_name = self.ctx.config.get('name', '')
        self.experiment_name = ctx.config.get('experiment/name', project_name)
        self.cluster_name = self.ctx.config.get('cluster/name', project_name)

        # get the preloaded workspace definition
        self.ws = Workspace.from_config()

    def _get_compute_target(self):
        compute_min_nodes = self.ctx.config.get('cluster/min_nodes',1)
        compute_max_nodes = self.ctx.config.get('cluster/max_nodes',4)
        compute_sku = self.ctx.config.get('cluster/type','STANDARD_D2_V2')
        compute_cluster = self.cluster_name

        if compute_cluster in self.ws.compute_targets:
            compute_target = self.ws.compute_targets[compute_cluster]
            if compute_target and type(compute_target) is AmlCompute:
                self.ctx.log(
                    'Found compute target. Just use it: ' + compute_cluster)
        else:
            self.ctx.log('Creating new AML compute context...')
            provisioning_config = AmlCompute.provisioning_configuration(
                vm_size=compute_sku, min_nodes=compute_min_nodes,
                max_nodes=compute_max_nodes)
            compute_target = ComputeTarget.create(
                self.ws, compute_cluster, provisioning_config)
            compute_target.wait_for_completion(show_output = True)
        return compute_target

    @error_handler
    def import_data(self):
        source = self.ctx.config.get('source', None)
        if source is None:
            raise AzureException('Please specify data source file...')
        ds = self.ws.get_default_datastore()
        ds.upload_files(files=[source], relative_root=None,
            target_path=None, overwrite=True, show_progress=True)
        dataset = os.path.basename(source)
        self.ctx.config.set('azure', 'dataset', dataset)
        self.ctx.config.write('azure')
        return {'dataset': dataset}

    @error_handler
    def train(self):
        config = self.ctx.config
        dataset = config.get('dataset', None)
        self.ctx.log("Starting search on %s Dataset..." % dataset)
        ds = self.ws.get_default_datastore()
        training_data = Dataset.Tabular.from_delimited_files(
            path=ds.path(dataset))
        #TODO training_data.drop_columns(columns)

        compute_target = self._get_compute_target()

        model_type = config.get('model_type')
        if (not model_type):
            raise AzureException('Please specify model type...')
        primary_metric = config.get('experiment/metric','spearman_correlation')
        if (not primary_metric):
            raise AzureException('Please specify primary metric...')
        #TODO: check if primary_metric is constent with model_type

        automl_settings = {
            #"name": "AutoML_Experiment_{0}".format(time.time()),
            "iteration_timeout_minutes" : config.get('experiment/iteration_timeout_minutes',10),
            "iterations" : config.get('experiment/max_n_trials',10),
            "primary_metric" : primary_metric,
            "verbosity" : logging.INFO,
            "n_cross_validations": config.get('experiment/cross_validation_folds',5),
            "enable_stack_ensemble": config.get('experiment/use_ensemble',False)
        }

        self.automl_config = AutoMLConfig(
            task=model_type,
            debug_log='automl_errors.log',
            path = os.getcwd(),
            compute_target = compute_target,
            training_data=training_data,
            label_column_name=config.get('target'),
            **automl_settings)

        experiment=Experiment(self.ws, self.experiment_name)
        remote_run = experiment.submit(self.automl_config, show_output=False)
        self.ctx.log("Started Experiment %s search..." % self.experiment_name)
        config.set('azure', 'experiment/name', self.experiment_name)
        config.set('azure', 'experiment/run_id', remote_run.run_id)
        config.write('azure')
        return {
            'eperiment_name': self.experiment_name,
            'run_id': remote_run.run_id}

    def _get_leaderboard(self, remote_run):
        primary_metric = remote_run.properties['primary_metric']
        children = list(remote_run.get_children(recursive=True))
        leaderboard = pd.DataFrame(index=['model id', 'algorithm', 'score'])
        goal_minimize = False
        for run in children:
            if('run_algorithm' in run.properties and 'score' in run.properties):
                if(run.properties['run_preprocessor']):
                    run_algorithm = '%s,%s' % (run.properties['run_preprocessor'],
                        run.properties['run_algorithm'])
                else:
                    run_algorithm = run.properties['run_algorithm']
                leaderboard[run.id] = [run.id,
                                      run_algorithm,
                                      float(run.properties['score'])]
                if('goal' in run.properties):
                    goal_minimize = run.properties['goal'].split('_')[-1] == 'min'

        leaderboard = leaderboard.T.sort_values(
            'score', ascending = goal_minimize)
        leaderboard = leaderboard.head(10)
        leaderboard.rename(columns={'score':primary_metric}, inplace=True)
        return leaderboard

    @error_handler
    def evaluate(self, run_id = None):
        if run_id is None:
            run_id = self.ctx.config.get('experiment/run_id', None)
        if run_id is None:
            raise AzureException(
                'Pleae provide Run ID (experiment/run_id) to evaluate')
        experiment = Experiment(self.ws, self.experiment_name)
        remote_run = AutoMLRun(experiment = experiment, run_id = run_id)
        leaderboard = self._get_leaderboard(remote_run).to_dict('records')
        print_table(self.ctx.log,leaderboard)
        status = remote_run.get_status()
        self.ctx.log('Status: %s' % status)
        return {
            'run_id': run_id,
            'leaderboard': leaderboard,
            'status': status }

    def _aci_service_name(self, model_name):
        return (model_name+'-service').lower()

    @error_handler
    def deploy(self, model_id, locally=False):
        if (locally):
            self.ctx.log('Local deployment step is not required for Azure..')
            return {'model_id': model_id}

        service_name = 'automl-remote'
        experiment_run_id = self.ctx.config.get(
            'experiment/run_id', None)

        experiment = Experiment(self.ws, self.experiment_name)
        remote_run = AutoMLRun(experiment = experiment, run_id = model_id)
        script_file_name = '.azureml/score_script.py'
        remote_run.download_file(
            'outputs/scoring_file_v_1_0_0.py', script_file_name)
        model_name = remote_run.properties['model_name']
        self.ctx.log('Model name: %s' % model_name)

        description = '%s-%s' % (service_name, model_name)
        experiment_run = AutoMLRun(
            experiment = experiment, run_id = experiment_run_id)
        # do we need to check for already registered one?
        model = experiment_run.register_model(
            model_name = model_name,
            description = description,
            tags = None)

        # It must only consist of lowercase letters, numbers, or dashes, start
        # with a letter, end with a letter or number, and be between 3 and 32
        # characters long.
        # TBD - service_name + suffix must satisfy requiremets
        aci_service_name = self._aci_service_name(model_name)
        self.ctx.log('Deploying AciWebservice %s ...' % aci_service_name)

        inference_config = InferenceConfig(
            environment = remote_run.get_environment(),
            entry_script = script_file_name)

        aciconfig = AciWebservice.deploy_configuration(
            cpu_cores = 1,
            memory_gb = 2,
            tags = {'type': "%s-inference-service" % service_name},
            description = "%s inference service" % service_name)

        # Remove any existing service under the same name.
        try:
            Webservice(self.ws, aci_service_name).delete()
            self.ctx.log('Remove any existing service under the same name...')
        except WebserviceException:
            pass

        aci_service = Model.deploy(
            self.ws,
            aci_service_name,
            [model],
            inference_config,
            aciconfig)
        aci_service.wait_for_deployment(True)
        self.ctx.log('%s state %s' % (aci_service_name, str(aci_service.state)))

        return {'model_id': model_id, 'aci_service_name': aci_service_name}

    def _save_predictions(self, df_predictions, filename):
        predicted_path = os.path.abspath(
            os.path.splitext(filename)[0] + "_predicted.csv")
        df_predictions.to_csv(predicted_path, index=False, encoding='utf-8')
        self.ctx.log('Predictions are saved to %s' % predicted_path)
        return predicted_path

    def _predict_local(self, predict_data, model_id, threshold):
        from auger.api.cloud.utils.dataframe import DataFrame

        experiment = Experiment(self.ws, self.experiment_name)
        run_id = model_id
        iteration = None
        parts = model_id.split('_')
        if len(parts) > 2:
            run_id = parts[0]+"_"+parts[1]
            iteration = parts[2]

        #print(run_id)
        #print(iteration)
        remote_run = AutoMLRun(experiment = experiment, run_id = run_id)
        best_run, fitted_model = remote_run.get_output(iteration=iteration)

        return fitted_model.predict(predict_data)

    @error_handler
    def predict(self, filename, model_id, threshold=None, locally=False):
        from auger.api.cloud.utils.dataframe import DataFrame
        target = self.ctx.config.get('target', None)
        predict_data = DataFrame.load(filename, target)

        y_pred = []
        if (locally):
            y_pred = self._predict_local(predict_data, model_id, threshold)
        else:
            input_payload = predict_data.to_json(orient='split', index=False)

            experiment = Experiment(self.ws, self.experiment_name)
            remote_run = AutoMLRun(experiment = experiment, run_id = model_id)
            model_name = remote_run.properties['model_name']

            aci_service_name = self._aci_service_name(model_name)
            aci_service = AciWebservice(self.ws, aci_service_name)
            input_payload = json.loads(input_payload)
            input_payload = {
                'method': 'predict',
                'data': input_payload['data']
            }
            input_payload = json.dumps(input_payload)

            response = aci_service.run(input_data = input_payload)
            y_pred = json.loads(response)['result']

        predict_data[target] = y_pred
        predicted = self._save_predictions(predict_data, filename)

        return {'predicted': predicted}

    def review(self):
        pass



import click

from a2ml.api.auger.auth import AugerAuth
from a2ml.api.utils.context import pass_context

class AuthCmd(object):

    def __init__(self, ctx):
        self.ctx = ctx

    def login(self, username, password, organization, url):
        AugerAuth(self.ctx).login(username, password, organization, url)

    def logout(self):
        AugerAuth(self.ctx).logout()

    def whoami(self):
        AugerAuth(self.ctx).whoami()

@click.group('auth', short_help='Authenticate with AutoML provider.')
@pass_context
def cmdl(ctx):
    """Authenticate with AutoML provider."""
    ctx.setup_logger(format='')

@click.command(short_help='Login to Auger.')
@click.option('--username', '-u', default=None, prompt='username',
    type=click.STRING, help='Auger username.')
@click.option('--organization', '-o', default=None, prompt='organization',
    type=click.STRING, help='Auger organization.')
@click.password_option('--password', '-p', prompt='password',
    confirmation_prompt=False, help='Auger password.')
@click.option('--system', '-s', default='production',
    type=click.Choice(['production','staging']),
    help='Auger API endpoint.')
@pass_context
def login(ctx, username, password, organization, system):
    """Login to Auger."""
    urls = {
        'production': 'https://app.auger.ai',
        'staging': 'https://app-staging.auger.ai'}
    AuthCmd(ctx).login(username, password, organization, urls[system])

@click.command(short_help='Logout from Auger.')
@pass_context
def logout(ctx):
    """Logout from Auger."""
    AuthCmd(ctx).logout()

@click.command(short_help='Display the current logged in user.')
@pass_context
def whoami(ctx):
    """Display the current logged in user."""
    ctx.setup_logger(format='')
    AuthCmd(ctx).whoami()


@pass_context
def add_commands(ctx):
    cmdl.add_command(login)
    cmdl.add_command(logout)
    cmdl.add_command(whoami)

add_commands()



from a2ml.api.utils.crud_runner import CRUDRunner
from a2ml.api.utils.error_handler import ErrorHandler

class A2MLExperiment(metaclass=ErrorHandler):

    def __init__(self, ctx, provider):
        super(A2MLExperiment, self).__init__()
        self.ctx = ctx
        self.runner = CRUDRunner(ctx, provider, 'experiment')

    def list(self):
        return self.runner.execute('list')

    def start(self):
        return self.runner.execute('start')

    def stop(self):
        return self.runner.execute('stop')

    def leaderboard(self, run_id):
        return self.runner.execute('leaderboard', run_id)

    def history(self):
        return self.runner.execute('history')





def error_handler(decorated):
    def wrapper(self, *args, **kwargs):
        try:
            return decorated(self, *args, **kwargs)
        except Exception as exc:
            if self.ctx.debug:
                import traceback
                traceback.print_exc()
            self.ctx.log(str(exc))
            raise exc
    return wrapper



import os
import sys
import click
import logging
from .config import Config

log = logging.getLogger("a2ml")

CONTEXT_SETTINGS = dict(auto_envvar_prefix='A2ML')
PROVIDERS = ['auger', 'google', 'azure']
PROVIDERS_META = '|'.join(PROVIDERS)


class Context(object):

    def __init__(self, name=''):
        super(Context, self).__init__()
        self.config = Config()
        if len(name) > 0:
            name = "{:<9}".format('[%s]' % name)
        self.name = name
        self.debug = self.config.get('debug', False)

    def get_providers(self, provider = None):
        if provider:
            providers = provider
        else:
            providers = self.config.get('providers', [])

        if isinstance(providers, (str,)):
            providers = [p.strip() for p in providers.split(',')]

        if isinstance(providers, (list,)):
            for p in providers:
                if p not in PROVIDERS:
                    raise Exception('Provider %s is not supported.' % p)
            return providers

        raise Exception('Expecting list of providers in config.yaml\providers')

    def copy(self, name):
        new = Context(name)
        new.config = Config(name)
        return new

    def log(self, msg, *args, **kwargs):
        log.info('%s%s' %(self.name, msg), *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        log.debug('%s%s' %(self.name, msg), *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        log.error('%s%s' %(self.name, msg), *args, **kwargs)

    @staticmethod
    def setup_logger(format='%(asctime)s %(name)s | %(message)s'):
        logging.basicConfig(
            stream=sys.stdout,
            datefmt='%H:%M:%S',
            format=format,
            level=logging.INFO)


pass_context = click.make_pass_decorator(Context, ensure=True)



from auger.api.cloud.rest_api import RestApi
from auger.api.credentials import Credentials
from auger.cli.commands.impl.experimentcmd import ExperimentCmd

class AugerExperiment(ExperimentCmd):

    def __init__(self, ctx):
        super(AugerExperiment, self).__init__(ctx)
        self.ctx.credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            self.ctx.credentials.api_url, self.ctx.credentials.token)



from a2ml.api.utils.error_handler import ErrorHandler
from a2ml.api.utils.provider_runner import ProviderRunner


class A2ML(metaclass=ErrorHandler):
    """Facade to A2ML providers."""

    def __init__(self, ctx, provider = None):
        super(A2ML, self).__init__()
        self.ctx = ctx
        self.runner = ProviderRunner(ctx, provider)
        self.showresult = False

    def import_data(self):
        results = self.runner.execute('import_data')
        if self.showresult:
            self.ctx.log(results)
        return results

    def train(self):
        results = self.runner.execute('train')
        if self.showresult:
            self.ctx.log(results)
        return results

    def evaluate(self, run_id = None):
        results = self.runner.execute('evaluate', run_id = run_id)
        if self.showresult:
            self.ctx.log(results)
        return results

    def deploy(self, model_id, locally=False):
        results = self.runner.execute('deploy', model_id, locally)
        if self.showresult:
            self.ctx.log(results)
        return results

    def predict(self, filename, model_id, threshold=None, locally=False):
        results = self.runner.execute('predict', filename, model_id, threshold, locally)
        if self.showresult:
            self.ctx.log(results)
        return results

    def review(self):
        results = self.runner.execute('review')
        if self.showresult:
            self.ctx.log(results)
        return results



import os
import logging

from auger.api.utils.config_yaml import ConfigYaml

log = logging.getLogger("a2ml")

class ConfigParts(object):
    parts = {}
    part_names = ['config', 'auger', 'azure', 'google']
    is_loaded = False

    @classmethod
    def load(cls, path = None):
        if path is None:
            path = os.getcwd()
        for pname in ConfigParts.part_names:
            filename = os.path.abspath(os.path.join(path, '%s.yaml' % pname))
            if os.path.isfile(filename):
                ConfigParts.parts[pname] = ConfigParts._load(filename)
        ConfigParts.is_loaded = True

    @classmethod
    def ismultipart(cls):
        return(len(ConfigParts.parts.keys()) > 1)

    @classmethod
    def keys(cls):
        return ConfigParts.parts.keys()

    @classmethod
    def part(cls, name):
        return ConfigParts.parts[name]

    @classmethod
    def _load(cls, name):
        part = ConfigYaml()
        if os.path.isfile(name):
            part.load_from_file(name)
        return part

class Config(object):

    def __init__(self, name = 'config'):
        super(Config, self).__init__()
        self.name = name
        self.path = None
        self.parts = ConfigParts()
        self.load()

    def get(self, path, default=None):
        if len(self.parts.keys()) == 0:
            return default
        if 'config' in self.parts.keys():
            default = self.parts.part('config').get(path, default)
        return self.parts.part(self.name).get(path, default)

    def set(self, part_name, path, value):
        if (part_name == 'config' and self.ismultipart()):
            self.parts.part('config').set(path, value)
        else:
            self.parts.part(self.name).set(path, value)

    def write(self, part_name):
        if (part_name == 'config' and self.ismultipart()):
            self.parts.part('config').write()
        else:
            self.parts.part(self.name).write()

    def ismultipart(self):
        return self.parts.ismultipart()

    def load(self, path = None, reload = False):
        if (not ConfigParts.is_loaded) or reload:
            ConfigParts.load(path)
        return self



from auger.api.cloud.rest_api import RestApi


def interceptor(payload, monkeypatch):
    def payloader(x, method, *args, **kwargs):
        return payload[method]
    monkeypatch.setattr(
        RestApi, 'call_ex', payloader)


def object_status_chain(statuses, monkeypatch):
    current = statuses.pop(0)
    if len(statuses):
        monkeypatch.setattr(
            RestApi, 'wait_for_object_status', lambda x, *a, **kw: current)
    return current



import os
import csv
from a2ml.api.auger.base import AugerBase
from auger.api.cloud.pipeline import AugerPipelineApi



class AugerAcutal(AugerBase):
    """Actual value for deployed Auger Pipeline."""

    def __init__(self, ctx):
        super(AugerPredict, self).__init__(ctx)

    @AugerBase._error_handler
    def actual(self, filename, model_id):
        # verify avalability of auger credentials
        self.credentials.verify()

        self.ctx.log('Sending actuals on data in %s' % filename)
        filename = os.path.abspath(filename)
        self._actual_to_cloud(filename, model_id)

    def _actual_to_cloud(self, filename, model_id):
        with open(filename) as f:
          a = [{k: v for k, v in row.items()} 
              for row in csv.DictReader(f, skipinitialspace=True)]

        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
        pipeline_api.actual(a)